! ==============================================================================================================================
! MODULE   : mod_test_par_netcdf 
!
! CONTACT      : orchidee-help _at_ listes.ipsl.fr
!
! LICENCE      : IPSL (2006)
! This software is governed by the CeCILL licence see ORCHIDEE/ORCHIDEE_CeCILL.LIC
!
!>\BRIEF      Test parallel netcdf 
!!
!!\n DESCRIPTION  : Basic parallel netcdf test. It will ensure the use of this libraries.
!!            The file was not properly formatted when written in Obelix.
!!            It took place by using the libraries netcdf/4p and hdf5/1.8.9p and 
!!            the use of 2 nodes and 4 core each.
!!            The file was not possible to be read afterwards.
!!
!!            This feature is used for the stomate_Cforcing_... files. Forcesoil
!!            was not able to read the input file.
!!
!! SVN              :
!! $HeadURL: svn://forge.ipsl.jussieu.fr/orchidee/branches/ORCHIDEE-MICT/ORCHIDEE/src_parallel/mod_orchidee_para.F90 $
!! $Date: 2017-10-26 15:35:04 +0200 (jeu. 26 oct. 2017) $
!! $Revision: 4719 $
!! \n
!_ ================================================================================================================================
MODULE mod_test_par_netcdf

  USE mpi
  USE netcdf
  USE ioipsl_para
  USE mod_orchidee_para_var
!  USE mod_orchidee_mpi_data
!  USE mod_orchidee_omp_data
!  USE mod_orchidee_transfert_para

  IMPLICIT NONE

  CHARACTER(LEN = *), PARAMETER :: FILE_NAME = "simple_xy_par.nc"
    
CONTAINS
    
  !!  =============================================================================================================================
  !! SUBROUTINE:  Test_parallel_netcdf
  !!
  !>\BRIEF	 Test for netcdf parallel 
  !!
  !! DESCRIPTION: This test is introduced because the writting of all
  !!            stomate_Cforcing_xxx.nc files was not properly done due to a bug in the libraries.
  !!            That was the case for Obelix with the following libraries:
  !!            - netcdf/4p and hdf5/1.8.9p
  !!            as well as the use of a job file with 2 nodes and 4 cores each.
  !!
  !!            Once the file was written, it was not possible to read.
  !!
  !!            It can be fixed by using the new libraries:
  !!            - netcdf/4.7 and hdf5/1.10.5
  !!
  !! \n
  !_ ============================================================================================================================== 
  SUBROUTINE Test_parallel_netcdf()
     INTEGER :: stat

     ! remove if it exists
     if (mpi_rank .EQ. 0) then
        open(unit=1234, iostat=stat, file=FILE_NAME, status='old')
        if (stat == 0) close(1234, status='delete')
     endif 

     CALL Test_parallel_write(FILE_NAME)
     CALL Test_parallel_read(FILE_NAME)

     ! remove if it exists
     if (mpi_rank .EQ. 0) then
        open(unit=1234, iostat=stat, file=FILE_NAME, status='old')
        if (stat == 0) close(1234, status='delete')
     endif

  END SUBROUTINE Test_parallel_netcdf

  !!  =============================================================================================================================
  !! SUBROUTINE:  Test_parallel_write
  !!
  !>\BRIEF	 This is a test for netcdf parallel write
  !!
  !! DESCRIPTION: 
  !!
  !!    Example from:
  !!  https://github.com/Unidata/netcdf-fortran/blob/master/examples/F90/simple_xy_par_wr2.f90
  !!
  !! \n
  !_ ============================================================================================================================== 
  SUBROUTINE Test_parallel_write(FILE_NAME)
     USE utils, ONLY: nccheck
     USE netcdf

     CHARACTER(LEN = *), INTENT(in) :: FILE_NAME 

     INTEGER, PARAMETER :: NDIMSWR = 3
     INTEGER :: ncid, varid, dimids(NDIMSWR)
     INTEGER :: x_dimid, y_dimid, t_dimid
     integer :: chunk_size(NDIMSWR)
     integer :: start(NDIMSWR), count(NDIMSWR)
     integer :: p, my_rank, ierr
     integer, allocatable :: data_out(:)
     integer :: x, stat

     allocate(data_out(mpi_size))
     do x = 1, mpi_size
       data_out(x) = mpi_rank
     end do

     ! Create the netCDF file. The NF90_NETCDF4 flag causes a
     ! HDF5/netCDF-4 file to be created. The comm and info parameters
     ! cause parallel I/O to be enabled. Use either NF90_MPIIO or
     ! NF90_MPIPOSIX to select between MPI/IO and MPI/POSIX.
     call nccheck(nf90_create(FILE_NAME, IOR(NF90_NETCDF4, NF90_MPIIO), ncid, &
             comm = MPI_COMM_WORLD, info = MPI_INFO_NULL))

     ! Define the dimensions. NetCDF will hand back an ID for
     ! each. Metadata operations must take place on all processors.
     call nccheck(nf90_def_dim(ncid, "x", mpi_size, x_dimid))
     call nccheck(nf90_def_dim(ncid, "y", mpi_size, y_dimid))
     call nccheck(nf90_def_dim(ncid, "t", NF90_UNLIMITED, t_dimid))

     ! The dimids array is used to pass the IDs of the dimensions of
     ! the variables. Note that in fortran arrays are stored in
     ! column-major format.
     dimids = (/ y_dimid, x_dimid, t_dimid /)

    ! define the chunk size (1 along unlimited time dimension)
    chunk_size = (/ mpi_size, 1, 1 /)

    ! Define the variable. The type of the variable in this case is
    ! NF90_INT (4-byte integer).
    call nccheck(nf90_def_var(ncid, "data", NF90_INT, dimids, varid, chunksizes=chunk_size))

    ! End define mode. This tells netCDF we are done defining
    ! metadata. This operation is collective and all processors will
    ! write their metadata to disk.
    call nccheck(nf90_enddef(ncid))

    ! Write the pretend data to the file. Each processor writes one row.
    start = (/ 1, mpi_rank + 1, 1/)
    count = (/ mpi_size, 1, 1 /)

    ! Unlimited dimensions require collective writes
    call nccheck(nf90_var_par_access(ncid, varid, nf90_collective))

    ! The unlimited axis prevents independent write tests
    ! Re-enable the rank test if independent writes are used in the future
    !if (my_rank.ne.0) &
    call nccheck(nf90_put_var(ncid, varid, data_out, start = start, &
            count = count))

    ! Close the file. This frees up any internal netCDF resources
    ! associated with the file, and flushes any buffers.
    call nccheck( nf90_close(ncid) )

    deallocate(data_out)

  END SUBROUTINE Test_parallel_write

  !!  =============================================================================================================================
  !! SUBROUTINE:  Test_parallel_read
  !!
  !>\BRIEF	 This is a test for netcdf parallel read 
  !!
  !! DESCRIPTION: 
  !!
  !!    Example from:
  !!  https://github.com/Unidata/netcdf-fortran/blob/master/examples/F90/simple_xy_par_rd.f90
  !!
  !! \n
  !_ ============================================================================================================================== 
  SUBROUTINE Test_parallel_read(FILE_NAME)
    USE utils, ONLY: nccheck
    USE netcdf

    CHARACTER(LEN = *), INTENT(in) :: FILE_NAME 

    INTEGER, PARAMETER :: NDIMS = 3
    INTEGER :: ncid, varid
    integer :: start(NDIMS), count(NDIMS)
    integer, allocatable :: data_in(:)
    integer :: x, stat

    allocate(data_in(mpi_size))

    ! Error is expected here
    call nc_check_test( nf90_open(FILE_NAME, IOR(NF90_NOWRITE, NF90_MPIIO), ncid, &
       comm = MPI_COMM_WORLD, info = MPI_INFO_NULL) )

    ! Get the varid of the data variable, based on its name.
    call nc_check_test( nf90_inq_varid(ncid, "data", varid) )

    ! Read the data.
    start = (/ 1, mpi_rank + 1, 1/)
    count = (/ mpi_size, 1, 1 /)
    call nc_check_test( nf90_get_var(ncid, varid, data_in, &
              start = start, count = count) )

    ! Check the data.
    do x = 1, mpi_size
       if (data_in(x) .ne. mpi_rank) then
          write(numout, *) "data_in(", x, ") = ", data_in(x)
          CALL ipslerr_p(3, 'Test_parallel_netcdf(Test_parallel_read)', "Failed!", &
                            'Different expected values found',  &
                            'in the netcdf file')
       endif
    end do

    call nc_check_test( nf90_close(ncid) )

    contains

  END SUBROUTINE Test_parallel_read

  !!  =============================================================================================================================
  !! SUBROUTINE: nc_check_test
  !!
  !>\BRIEF	    Check netcdf output
  !!
  !! DESCRIPTION: 
  !!
  !! \n
  !_ ============================================================================================================================== 
  subroutine nc_check_test(status)
     integer, intent(in) :: status

     IF(status /= nf90_noerr) THEN
        print *, "nc_check_test:: Netcdf error: ", nf90_strerror(status)
        write (numout,*) "nc_check_test:: Netcdf error: ", nf90_strerror(status)

        CALL ipslerr_p(3, 'Test_parallel_netcdf(nc_check_test)', "Failed!", 'Error code:', trim(nf90_strerror(status)))
     ENDIF 

  end subroutine nc_check_test
  
END MODULE mod_test_par_netcdf
